@@ -43,6 +43,7 @@ import Data.Text (Text, pack)
 import qualified Data.Text.Lazy as TL
 import Text.Blaze (Html, preEscapedText, toHtml)
 import qualified Data.Foldable as F
+import Control.Monad (mplus)
 
 type Render url = url -> [(Text, Text)] -> Text
 type Translate msg = msg -> Html
@@ -61,31 +62,43 @@ docsToExp env hr scope docs = do
         [x] -> return x
         _ -> return $ DoE $ map NoBindS exps
 
+unIdent :: Ident -> String
+unIdent (Ident s) = s
+
 docToExp :: Env -> HamletRules -> Scope -> Doc -> Q Exp
-docToExp env hr scope (DocForall list ident@(Ident name) inside) = do
+docToExp env hr scope (DocForall list idents inside) = do
     let list' = derefToExp scope list
-    name' <- newName name
-    let scope' = (ident, VarE name') : scope
+    names <- mapM (newName . unIdent) idents
+    let pairs = zip idents (map VarE names)
+    let scope' = pairs ++ scope
     mh <- [|F.mapM_|]
     inside' <- docsToExp env hr scope' inside
-    let lam = LamE [VarP name'] inside'
+    let lam = flip LamE inside' $ case names of
+            [x] -> [VarP x]
+            _ -> [TupP $ map VarP names]
     return $ mh `AppE` lam `AppE` list'
 docToExp env hr scope (DocWith [] inside) = do
     inside' <- docsToExp env hr scope inside
     return $ inside'
-docToExp env hr scope (DocWith ((deref,ident@(Ident name)):dis) inside) = do
+docToExp env hr scope (DocWith ((deref, idents):dis) inside) = do
     let deref' = derefToExp scope deref
-    name' <- newName name
-    let scope' = (ident, VarE name') : scope
+    names <- mapM (newName . unIdent) idents
+    let pairs = zip idents (map VarE names)
+    let scope' = pairs ++ scope
     inside' <- docToExp env hr scope' (DocWith dis inside)
-    let lam = LamE [VarP name'] inside'
+    let lam = flip LamE inside' $ case names of
+            [x] -> [VarP x]
+            _ -> [TupP $ map VarP names]
     return $ lam `AppE` deref'
-docToExp env hr scope (DocMaybe val ident@(Ident name) inside mno) = do
+docToExp env hr scope (DocMaybe val idents inside mno) = do
     let val' = derefToExp scope val
-    name' <- newName name
-    let scope' = (ident, VarE name') : scope
+    names <- mapM (newName . unIdent) idents
+    let pairs = zip idents (map VarE names)
+    let scope' = pairs ++ scope
     inside' <- docsToExp env hr scope' inside
-    let inside'' = LamE [VarP name'] inside'
+    let inside'' = flip LamE inside' $ case names of
+            [x] -> [VarP x]
+            _ -> [TupP $ map VarP names]
     ninside' <- case mno of
                     Nothing -> [|Nothing|]
                     Just no -> do
@@ -110,6 +123,18 @@ docToExp env hr scope (DocCond conds final) = do
         let d' = derefToExp scope d
         docs' <- docsToExp env hr scope docs
         return $ TupE [d', docs']
+docToExp env hr scope (DocCase deref cases) = do
+    let exp_ = derefToExp scope deref
+    matches <- mapM toMatch cases
+    return $ CaseE exp_ matches
+  where
+    toMatch (idents, inside) = do
+        let pat = case map unIdent idents of
+                    ["_"] -> WildP
+                    strs -> let (constr:fields) = map mkName strs
+                            in ConP constr (map VarP fields)
+        insideExp <- docsToExp env hr scope inside
+        return $ Match pat (NormalB insideExp) []
 docToExp env hr v (DocContent c) = contentToExp env hr v c
 
 contentToExp :: Env -> HamletRules -> Scope -> Content -> Q Exp
@@ -253,14 +278,9 @@ strToExp "" = error "strToExp on empty string"
 -- first is performed. In there are no true values, then the second argument is
 -- performed, if supplied.
 condH :: Monad m => [(Bool, m ())] -> Maybe (m ()) -> m ()
-condH [] Nothing = return ()
-condH [] (Just x) = x
-condH ((True, y):_) _ = y
-condH ((False, _):rest) z = condH rest z
+condH bms mm = fromMaybe (return ()) $ lookup True bms `mplus` mm
 
 -- | Runs the second argument with the value in the first, if available.
 -- Otherwise, runs the third argument, if available.
 maybeH :: Monad m => Maybe v -> (v -> m ()) -> Maybe (m ()) -> m ()
-maybeH Nothing _ Nothing = return ()
-maybeH Nothing _ (Just x) = x
-maybeH (Just v) f _ = f v
+maybeH mv f mm = fromMaybe (return ()) $ fmap f mv `mplus` mm
