@@ -43,13 +43,15 @@ data Content = ContentRaw String
              | ContentMsg Deref
     deriving (Show, Eq, Read, Data, Typeable)
 
-data Line = LineForall Deref Ident
+data Line = LineForall Deref [Ident]
           | LineIf Deref
           | LineElseIf Deref
           | LineElse
-          | LineWith [(Deref, Ident)]
-          | LineMaybe Deref Ident
+          | LineWith [(Deref, [Ident])]
+          | LineMaybe Deref [Ident]
           | LineNothing
+          | LineCase Deref
+          | LineOf [Ident]
           | LineTag
             { _lineTagName :: String
             , _lineAttr :: [(Maybe Deref, String, [Content])]
@@ -70,6 +72,7 @@ parseLine set = do
     ss <- fmap sum $ many ((char ' ' >> return 1) <|>
                            (char '\t' >> fail "Tabs are not allowed in Hamlet indentation"))
     x <- doctype <|>
+         doctypeDollar <|>
          comment <|>
          htmlComment <|>
          backslash <|>
@@ -80,7 +83,10 @@ parseLine set = do
          (try (string "$nothing") >> spaceTabs >> eol >> return LineNothing) <|>
          controlForall <|>
          controlWith <|>
+         controlCase <|>
+         controlOf <|>
          angle <|>
+         invalidDollar <|>
          (eol' >> return (LineContent [])) <|>
          (do
             cs <- content InContent
@@ -96,6 +102,17 @@ parseLine set = do
     doctype = do
         try $ string "!!!" >> eol
         return $ LineContent [ContentRaw $ hamletDoctype set ++ "\n"]
+    doctypeDollar = do
+        _ <- try $ string "$doctype "
+        name <- many $ noneOf "\r\n"
+        eol
+        case lookup name doctypeNames of
+            Nothing -> fail $ "Unknown doctype name: " ++ name
+            Just val -> return $ LineContent [ContentRaw $ val ++ "\n"]
+
+    invalidDollar = do
+        _ <- char '$'
+        fail $ "Received a command I did not understand. If you wanted a literal $, start the line with a backslash."
     comment = do
         _ <- try $ string "$#"
         _ <- many $ noneOf "\r\n"
@@ -132,7 +149,7 @@ parseLine set = do
         eol
         return $ LineElseIf x
     binding = do
-        y <- ident
+        y <- identPattern
         spaces
         _ <- string "<-"
         spaces
@@ -157,6 +174,19 @@ parseLine set = do
         spaces
         bindings <- (binding `sepBy` bindingSep) `endBy` eol
         return $ LineWith $ concat bindings -- concat because endBy returns a [[(Deref,Ident)]]
+    controlCase = do
+        _ <- try $ string "$case"
+        spaces
+        x <- parseDeref
+        _ <- spaceTabs
+        eol
+        return $ LineCase x
+    controlOf = do
+        _   <- try $ string "$of"
+        pat <- many1 $ try $ spaces >> ident
+        _   <- spaceTabs
+        eol
+        return $ LineOf pat
     content cr = do
         x <- many $ content' cr
         case cr of
@@ -201,24 +231,28 @@ parseLine set = do
         content cr
     tagIdent = char '#' >> TagIdent <$> tagAttribValue NotInQuotes
     tagCond = do
-        _ <- char ':'
-        d <- parseDeref
-        _ <- char ':'
+        d <- between (char ':') (char ':') parseDeref
         tagClass (Just d) <|> tagAttrib (Just d)
     tagClass x = char '.' >> (TagClass . (,) x) <$> tagAttribValue NotInQuotes
     tagAttrib cond = do
         s <- many1 $ noneOf " \t=\r\n>"
-        v <- (do
-            _ <- char '='
-            s' <- tagAttribValue NotInQuotesAttr
-            return s') <|> return []
+        v <- (char '=' >> tagAttribValue NotInQuotesAttr) <|> return []
         return $ TagAttrib (cond, s, v)
     tag' = foldr tag'' ("div", [], [])
     tag'' (TagName s) (_, y, z) = (s, y, z)
     tag'' (TagIdent s) (x, y, z) = (x, (Nothing, "id", s) : y, z)
     tag'' (TagClass s) (x, y, z) = (x, y, s : z)
     tag'' (TagAttrib s) (x, y, z) = (x, s : y, z)
+
+    ident :: Parser Ident
     ident = Ident <$> many1 (alphaNum <|> char '_' <|> char '\'')
+
+    identPattern :: Parser [Ident]
+    identPattern = (between
+        (char '(' >> spaces)
+        (spaces >> char ')' >> spaces)
+        (sepBy1 ident (spaces >> char ',' >> spaces))
+        ) <|> (return <$> ident)
     angle = do
         _ <- char '<'
         name' <- many  $ noneOf " \t.#\r\n!>"
@@ -226,10 +260,7 @@ parseLine set = do
         xs <- many $ try ((many $ oneOf " \t") >>
               (tagIdent <|> tagCond <|> tagClass Nothing <|> tagAttrib Nothing))
         _ <- many $ oneOf " \t"
-        c <- (eol >> return []) <|> (do
-            _ <- char '>'
-            c <- content InContent
-            return c)
+        c <- (eol >> return []) <|> (char '>' >> content InContent)
         let (tn, attr, classes) = tag' $ TagName name : xs
         return $ LineTag tn attr c classes
 
@@ -249,10 +280,11 @@ nestLines ((i, l):rest) =
     let (deeper, rest') = span (\(i', _) -> i' > i) rest
      in Nest l (nestLines deeper) : nestLines rest'
 
-data Doc = DocForall Deref Ident [Doc]
-         | DocWith [(Deref,Ident)] [Doc]
+data Doc = DocForall Deref [Ident] [Doc]
+         | DocWith [(Deref, [Ident])] [Doc]
          | DocCond [(Deref, [Doc])] (Maybe [Doc])
-         | DocMaybe Deref Ident [Doc] (Maybe [Doc])
+         | DocMaybe Deref [Ident] [Doc] (Maybe [Doc])
+         | DocCase Deref [([Ident], [Doc])]
          | DocContent Content
     deriving (Show, Eq, Read, Data, Typeable)
 
@@ -281,6 +313,14 @@ nestToDoc set (Nest (LineMaybe d i) inside:rest) = do
             _ -> return (Nothing, rest)
     rest'' <- nestToDoc set rest'
     Ok $ DocMaybe d i inside' nothing : rest''
+nestToDoc set (Nest (LineCase d) inside:rest) = do
+    let getOf (Nest (LineOf pat) insideC) = do
+            insideC' <- nestToDoc set insideC
+            Ok (pat, insideC')
+        getOf _ = Error "Inside a $case there may only be $of.  Use '$of _' for a wildcard."
+    cases <- mapM getOf inside
+    rest' <- nestToDoc set rest
+    Ok $ DocCase d cases : rest'
 nestToDoc set (Nest (LineTag tn attrs content classes) inside:rest) = do
     let attrFix (x, y, z) = (x, y, [(Nothing, z)])
     let takeClass (a, "class", b) = Just (a, b)
@@ -325,6 +365,7 @@ nestToDoc set (Nest (LineContent content) inside:rest) = do
 nestToDoc _set (Nest (LineElseIf _) _:_) = Error "Unexpected elseif"
 nestToDoc _set (Nest LineElse _:_) = Error "Unexpected else"
 nestToDoc _set (Nest LineNothing _:_) = Error "Unexpected nothing"
+nestToDoc _set (Nest (LineOf _) _:_) = Error "Unexpected 'of' (did you forget a $case?)"
 
 compressDoc :: [Doc] -> [Doc]
 compressDoc [] = []
@@ -340,6 +381,8 @@ compressDoc (DocCond [(a, x)] Nothing:DocCond [(b, y)] Nothing:rest)
 compressDoc (DocCond x y:rest) =
     DocCond (map (second compressDoc) x) (compressDoc `fmap` y)
     : compressDoc rest
+compressDoc (DocCase d cs:rest) =
+    DocCase d (map (second compressDoc) cs) : compressDoc rest
 compressDoc (DocContent (ContentRaw ""):rest) = compressDoc rest
 compressDoc ( DocContent (ContentRaw x)
             : DocContent (ContentRaw y)
@@ -367,7 +410,7 @@ attrToContent (Nothing, k, [(Nothing, v)]) =
   ++ [DocContent $ ContentRaw "\""]
 attrToContent (Nothing, k, v) = -- only for class
       DocContent (ContentRaw (' ' : k ++ "=\""))
-    : concat (map go $ init v)
+    : concatMap go (init v)
     ++ go' (last v)
     ++ [DocContent $ ContentRaw "\""]
   where
@@ -450,8 +493,16 @@ parseConds :: HamletSettings
            -> Result ([(Deref, [Doc])], Maybe [Doc], [Nest])
 parseConds set front (Nest LineElse inside:rest) = do
     inside' <- nestToDoc set inside
-    Ok $ (front [], Just inside', rest)
+    Ok (front [], Just inside', rest)
 parseConds set front (Nest (LineElseIf d) inside:rest) = do
     inside' <- nestToDoc set inside
     parseConds set (front . (:) (d, inside')) rest
 parseConds _ front rest = Ok (front [], Nothing, rest)
+
+doctypeNames :: [(String, String)]
+doctypeNames =
+    [ ("5", "<!DOCTYPE html>")
+    , ("html", "<!DOCTYPE html>")
+    , ("1.1", "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">")
+    , ("strict", "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">")
+    ]
