@@ -60,15 +60,17 @@ parseHamletRT set s =
         Error s' -> failure $ HamletParseException s'
         Ok x -> liftM HamletRT $ mapM convert x
   where
-    convert x@(DocForall deref (Ident ident) docs) = do
+    convert x@(DocForall deref [Ident ident] docs) = do
         deref' <- flattenDeref' x deref
         docs' <- mapM convert docs
         return $ SDForall deref' ident docs'
-    convert x@(DocMaybe deref (Ident ident) jdocs ndocs) = do
+    convert DocForall{} = error "Runtime Hamlet does not currently support tuple patterns"
+    convert x@(DocMaybe deref [Ident ident] jdocs ndocs) = do
         deref' <- flattenDeref' x deref
         jdocs' <- mapM convert jdocs
         ndocs' <- maybe (return []) (mapM convert) ndocs
         return $ SDMaybe deref' ident jdocs' ndocs'
+    convert DocMaybe{} = error "Runtime Hamlet does not currently support tuple patterns"
     convert (DocContent (ContentRaw s')) = return $ SDRaw s'
     convert x@(DocContent (ContentVar deref)) = do
         y <- flattenDeref' x deref
@@ -91,6 +93,7 @@ parseHamletRT set s =
             docs'' <- mapM convert docs'
             return (deref', docs'')
     convert DocWith{} = error "Runtime hamlet does not currently support $with"
+    convert DocCase{} = error "Runtime hamlet does not currently support $case"
 
 renderHamletRT :: Failure HamletException m
                => HamletRT
